AF_INET
PF_INET
AF_INET6
PF_INET6
AF_LINK
PF_LINK
AF_LOCAL
PF_LOCAL
AF_UNIX
PF_UNIX
AF_MAX
AF_UNSPEC
PF_UNSPEC
AF_ROUTE
PF_ROUTE

AI_CANONNAME
AI_FQDN
AI_NUMERICHOST
AI_NUMERICSERV
AI_PASSIVE

IP_ADD_MEMBERSHIP
IP_ADD_SOURCE_MEMBERSHIP
IP_BLOCK_SOURCE
IP_DROP_MEMBERSHIP
IP_DROP_SOURCE_MEMBERSHIP
IP_FREEBIND
IP_HDRINCL
IP_IPSEC_POLICY
IP_MINTTL
IP_MSFILTER
IP_MTU
IP_MTU_DISCOVER
IP_MULTICAST_ALL
IP_MULTICAST_IF
IP_MULTICAST_LOOP
IP_MULTICAST_TTL
IP_OPTIONS
IP_ORIGDSTADDR
IP_PASSSEC
IP_PKTINFO
IP_PKTOPTIONS
IP_PMTUDISC_DO
IP_PMTUDISC_DONT
IP_PMTUDISC_PROBE
IP_PMTUDISC_WANT
IP_RECVDSTADDR
IP_RECVERR
IP_RECVOPTS
IP_RECVORIGDSTADDR
IP_RECVRETOPTS
IP_RECVTOS
IP_RECVTTL
IP_RETOPTS
IP_ROUTER_ALERT
IP_TOS
IP_TRANSPARENT
IP_TTL
IP_UNBLOCK_SOURCE
IP_XFRM_POLICY

IPV6_JOIN_GROUP
IPV6_LEAVE_GROUP
IPV6_MULTICAST_HOPS
IPV6_MULTICAST_IF
IPV6_MULTICAST_LOOP
IPV6_UNICAST_HOPS
IPV6_V6ONLY

IPPROTO_AH
IPPROTO_DSTOPTS
IPPROTO_ESP
IPPROTO_FRAGMENT
IPPROTO_ICMP
IPPROTO_ICMPV6
IPPROTO_IP
IPPROTO_IPV6
IPPROTO_NONE
IPPROTO_RAW
IPPROTO_ROUTING
IPPROTO_TCP
IPPROTO_UDP

MCAST_BLOCK_SOURCE
MCAST_JOIN_GROUP
MCAST_JOIN_SOURCE_GROUP
MCAST_LEAVE_GROUP
MCAST_LEAVE_SOURCE_GROUP
MCAST_MSFILTER
MCAST_UNBLOCK_SOURCE

MSG_BCAST
MSG_CTRUNC
MSG_DONTROUTE
MSG_DONTWAIT
MSG_EOR
MSG_MCAST
MSG_NOSIGNAL
MSG_OOB
MSG_PEEK
MSG_TRUNC
MSG_WAITALL

NI_DGRAM
NI_MAXHOST
NI_MAXSERV
NI_NAMEREQD
NI_NOFQDN
NI_NUMERICHOST
NI_NUMERICSERV

SHUT_RD
SHUT_WR
SHUT_RDWR

SO_BINDANY
SO_BROADCAST
SO_DEBUG
SO_DONTROUTE
SO_ERROR
SO_KEEPALIVE
SO_LINGER
SO_NOSIGPIPE
SO_OOBINLINE
SO_PEERCRED
SO_RCVBUF
SO_RCVLOWAT
SO_RCVTIMEO
SO_REUSEADDR
SO_REUSEPORT
SO_RTABLE
SO_SNDBUF
SO_SNDLOWAT
SO_SNDTIMEO
SO_SPLICE
SO_TIMESTAMP
SO_TYPE

SOCK_DGRAM
SOCK_RAW
SOCK_SEQPACKET
SOCK_STREAM

SOL_SOCKET

TCP_CONGCTL
TCP_CONGESTION
TCP_CORK
TCP_DEFER_ACCEPT
TCP_INFO
TCP_KEEPCNT
TCP_KEEPIDLE
TCP_KEEPINIT
TCP_KEEPINTVL
TCP_LINGER2
TCP_MAXSEG
TCP_MD5SIG
TCP_NODELAY
TCP_QUICKACK
TCP_SACK_ENABLE
TCP_SYNCNT
TCP_WINDOW_CLAMP
