/*
** mrdberror.h - mruby debugger error code
**
*/

#ifndef MRDBERROR_H
#define MRDBERROR_H

#define MRB_DEBUG_OK                   (0)
#define MRB_DEBUG_NOBUF                (-1)
#define MRB_DEBUG_INVALID_ARGUMENT     (-2)

#define MRB_DEBUG_BREAK_INVALID_LINENO (-11)
#define MRB_DEBUG_BREAK_INVALID_FILE   (-12)
#define MRB_DEBUG_BREAK_INVALID_NO     (-13)
#define MRB_DEBUG_BREAK_NUM_OVER       (-14)
#define MRB_DEBUG_BREAK_NO_OVER        (-15)

#endif

